from pprint import pprint
from subprocess import CalledProcessError, run as sp_run
from typing import Optional
import sys
import re

from distlib.version import NormalizedVersion  # type: ignore


class CommandRunner:

    def __init__(self, verbose=False, dry=False):
        self.verbose = verbose
        self.dry = dry

    def run(self, *args, **kwargs):
        """Wrapper around subprocess.run that optionally prints the command."""
        if self.verbose:
            pprint([*args])
            pprint(kwargs)
        elif self.dry:
            from shlex import join
            print(join(args[0]))
        if not self.dry:
            return sp_run(*args, **kwargs)

    def check_program_version(
        self,
        program: str,
        minimum_version: Optional[NormalizedVersion],
        name: Optional[str],
        check_version: bool = True,
    ):
        """Check if there's a new enough version of the given command available
        in PATH."""
        name = name or program
        try:
            # Try running the command
            cmd = [program, '--version'] if check_version else [program, '-h']
            res = self.run(cmd,
                           check=True,
                           capture_output=True,
                           encoding='utf-8')
            # Try finding the version
            if res is not None and check_version:
                m = re.search(r'\d+(\.\d+){1,}', res.stdout)
                if not m:
                    raise RuntimeError(f"Unexpected {name} version output")
                program_version = NormalizedVersion(m.group(0))
                if self.verbose: print("Found", name, program_version)
                # Check if the version is new enough
                if minimum_version is not None:
                    if program_version < minimum_version:
                        raise RuntimeError(f"{name} too old")
        except CalledProcessError as e:
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
                print(e.stdout)
                print(e.stderr, file=sys.stderr)
            return False
        except Exception as e:
            # If any of that failed, return False
            if self.verbose:
                print(f'{type(e).__module__}.{type(e).__name__}', e, sep=': ')
            return False
        return True



from . import __version__
from pathlib import Path
import click


def cmake_command(directory, build_path, verbose, dry, native, cross, local):

    def get_cmaker():
        from .build import _BuildBackend as backend
        from .datastructures import PackageInfo
        from .cmd_runner import CommandRunner
        source_dir = Path(directory or '.').resolve()
        config_settings = {
            "--cross": cross,
            "--local": local,
        }
        # Read configuration and package metadata
        cfg, pkg, metadata = backend.read_all_metadata(source_dir,
                                                       config_settings,
                                                       verbose)

        if not cfg.cmake:
            print("Not a CMake package")
            return None

        pkg_info = PackageInfo(
            version=metadata.version,
            package_name=cfg.package_name,
            module_name=cfg.module["name"],
        )
        # Select the right configuration (native build or cross build)
        cmake_cfg, native_cmake_cfg = backend.get_cmake_configs(cfg)
        cmake_cfg = native_cmake_cfg if native else cmake_cfg
        cross_cfg = None if native else cfg.cross
        # Override the build folder
        if build_path is not None:
            cmake_cfg['build_path'] = str(build_path)
        # Configure all CMake options
        return backend.get_cmaker(pkg_dir=source_dir,
                                  install_dir=None,
                                  cmake_cfg=cmake_cfg,
                                  cross_cfg=cross_cfg,
                                  native_install_dir=None,
                                  package_info=pkg_info,
                                  runner=CommandRunner(verbose=verbose,
                                                       dry=dry))

    return get_cmaker


@click.group()
@click.option("-C",
              "--directory",
              type=click.Path(exists=True, file_okay=False, dir_okay=True),
              required=False,
              help="The directory containing pyproject.toml.")
@click.option("-B",
              "--build-path",
              type=click.Path(exists=False, file_okay=False, dir_okay=True),
              required=False,
              default=None,
              help="Override py-build-cmake's default build folder.")
@click.option("-V",
              "--verbose",
              is_flag=True,
              help="Print verbose information about the commands being "
              "executed.")
@click.option("-n",
              "--dry",
              is_flag=True,
              help="Print the commands without actually invoking CMake.")
@click.option("--native",
              is_flag=True,
              help="When the configuration requests cross-compiling, "
              "configure for a native build instead. Has no effect otherwise.")
@click.option("--local",
              type=click.Path(exists=False, file_okay=True, dir_okay=False),
              required=False,
              help="Specifies a toml file that overrides the "
              "tool.py-build-cmake section of pyproject.toml, "
              "similar to py-build-cmake.local.toml.")
@click.option("--cross",
              type=click.Path(exists=False, file_okay=True, dir_okay=False),
              required=False,
              help="Specifies a toml file that overrides the "
              "tool.py-build-cmake.cross section of pyproject.toml, "
              "similar to py-build-cmake.cross.toml.")
@click.version_option(__version__, "-v", "--version")
@click.pass_context
def cli(ctx: click.Context, **kwargs):
    ctx.obj = cmake_command(**kwargs)


@cli.command(help="Configure the CMake project.")
@click.pass_obj
@click.option("--preset", nargs=1, type=str, required=False, metavar="PRESET")
@click.argument("args", nargs=-1, required=False)
def configure(obj, preset, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.conf_settings.args += args or []
    if preset is not None:
        cmaker.conf_settings.preset = preset
    cmaker.configure()


@cli.command(help="Build the CMake project.")
@click.pass_obj
@click.option("--preset",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="PRESET")
@click.option("--config",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="CONFIG")
@click.argument("args", nargs=-1, required=False)
def build(obj, preset, config, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.build_settings.args += args or []
    if preset or config:
        cmaker.build_settings.presets = []
        cmaker.build_settings.configs = []
    if preset:
        cmaker.build_settings.presets = preset
    if config:
        cmaker.build_settings.configs = config
    cmaker.build()


@cli.command(help="Install the CMake project.")
@click.pass_obj
@click.option("--preset",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="PRESET")
@click.option("--config",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="CONFIG")
@click.option("--component",
              nargs=1,
              multiple=True,
              type=str,
              required=False,
              metavar="COMP")
@click.argument("args", nargs=-1, required=False)
def install(obj, preset, config, component, args):
    cmaker = obj()
    if cmaker is None:
        return
    cmaker.install_settings.args += args or []
    if preset or config:
        cmaker.install_settings.presets = []
        cmaker.install_settings.configs = []
    if preset:
        cmaker.install_settings.presets = preset
    if config:
        cmaker.install_settings.configs = config
    if component:
        cmaker.install_settings.components = component
    cmaker.install()


@cli.group(help="Config file operations.")
def config():
    pass


@config.command(help="Print documentation for the config file format.")
@click.option("--md", is_flag=True, help="Use the MarkDown format.")
@click.option("--component",
              is_flag=True,
              help="Documentation for the build_component backend.")
def format(md, component):
    from .pyproject_options import get_options, get_component_options
    from .config_options import pth
    if md:
        from .help import help_print_md as help_print
    else:
        from .help import help_print as help_print
    help_pth = pth('pyproject.toml/tool/py-build-cmake')
    pr_md = print if md else lambda *args, **kwargs: None
    pr_tx = lambda *args, **kwargs: None if md else print
    if component:
        pr_tx("List of py-build-cmake pyproject.toml options for the "
              "build_component backend:")
        pr_md("# py-build-cmake component build backend\n"
              "The `py_build_cmake.build_component` build backend allows "
              "building packages containing additional binaries that are not "
              "included with the main distribution.\n")
        pr_md("A possible use case is distributing debug symbols: these files "
              "can be large, and most users don't need them, so distributing "
              "them in a separate package makes sense.\n")
        pr_md(
            "You can find an example in the "
            "[alpaqa](https://pypi.org/project/alpaqa) package: "
            "<https://github.com/kul-optec/alpaqa/blob/main/python/alpaqa-debug/pyproject.toml>"
        )
        help_print(get_component_options(Path('/'))[help_pth])

    else:
        pr_tx("List of py-build-cmake pyproject.toml options:")
        pr_md("# py-build-cmake configuration options\n")
        pr_md("These options go in the `[tool.py-build-cmake]` section of "
              "the `pyproject.toml` configuration file.\n")
        help_print(get_options(Path('/'))[help_pth])
        pr_md("# Local overrides\n")
        pr_md("Additionally, two extra configuration files can be placed in "
              "the same directory as `pyproject.toml` to override some "
              "options for your specific use case:\n\n"
              "- `py-build-cmake.local.toml`: the options in this file "
              "override the values in the `tool.py-build-cmake` section of "
              "`pyproject.toml`.<br/>This is useful if you need specific "
              "arguments or CMake options to compile the package on your "
              "system.\n"
              "- `py-build-cmake.cross.toml`: the options in this file "
              "override the values in the `tool.py-build-cmake.cross` section "
              "of `pyproject.toml`.<br/>Useful for cross-compiling the "
              "package without having to edit the main configuration file.\n")
        pr_md("# Command line overrides\n")
        pr_md("Instead of using the `py-build-cmake.local.toml` and "
              "`py-build-cmake.cross.toml` files, you can also include "
              "additional config files using command line options:\n\n"
              "- `--local`: specifies a toml file that overrides the "
              "`tool.py-build-cmake` section of `pyproject.toml`, "
              "similar to `py-build-cmake.local.toml`\n"
              "- `--cross`: specifies a toml file that overrides the "
              "`tool.py-build-cmake.cross` section of `pyproject.toml`, "
              "similar to `py-build-cmake.cross.toml`\n\n"
              "These command line overrides are applied after the "
              "`py-build-cmake.local.toml` and `py-build-cmake.cross.toml` "
              "files in the project folder (if any).\n\n"
              "When using PyPA build, these flags can be specified using "
              "the `-C` or `--config-setting` flag: \n"
              "```sh\n"
              "python -m build . -C--cross=/path/to/my-cross-config.toml\n"
              "```\n"
              "The same flag may appear multiple times, for example: \n"
              "```sh\n"
              "python -m build . -C--local=conf-A.toml -C--local=conf-B.toml\n"
              "```\n"
              "For PyPA pip, you can use the `--config-settings` flag "
              "instead.")


if __name__ == '__main__':
    cli()



import tempfile
from pathlib import Path
from typing import Any, Dict, List, Optional
import os
from pprint import pprint

from . import config
from .build import _BuildBackend
from .datastructures import BuildPaths, PackageInfo
from .cmd_runner import CommandRunner

import flit_core.config


class _BuildComponentBackend(object):

    # --- Constructor ---------------------------------------------------------

    def __init__(self) -> None:
        self.runner: CommandRunner = CommandRunner()

    @property
    def verbose(self):
        return self.runner.verbose

    # --- Methods required by PEP 517 -----------------------------------------

    def get_requires_for_build_wheel(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-wheel"""
        self.parse_config_settings(config_settings)

        comp_pyproject = Path('pyproject.toml').resolve()
        comp_cfg = self.read_component_config(comp_pyproject, config_settings,
                                              self.verbose)
        cfg = _BuildBackend.read_config(
            Path(comp_cfg.component['main_project']) / 'pyproject.toml',
            config_settings, self.verbose)
        return _BuildBackend.get_requires_build_project(
            config_settings, cfg, self.runner)

    def get_requires_for_build_editable(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0660/#get-requires-for-build-editable"""
        return self.get_requires_for_build_wheel(config_settings)

    def get_requires_for_build_sdist(self, config_settings=None):
        """https://www.python.org/dev/peps/pep-0517/#get-requires-for-build-sdist"""
        return []

    def build_wheel(self,
                    wheel_directory,
                    config_settings=None,
                    metadata_directory=None):
        """https://www.python.org/dev/peps/pep-0517/#build-wheel"""
        assert metadata_directory is None

        # Parse options
        self.parse_config_settings(config_settings)

        # Build wheel
        with tempfile.TemporaryDirectory() as tmp_build_dir:
            whl_name = self.build_wheel_in_dir(wheel_directory, tmp_build_dir,
                                               config_settings)
        return whl_name

    def build_editable(self,
                       wheel_directory,
                       config_settings=None,
                       metadata_directory=None):
        raise NotImplementedError(
            "Editable installation not supported for individual components.")

    def build_sdist(self, sdist_directory, config_settings=None):
        raise NotImplementedError(
            "Source distribution not supported for individual components.")

    # --- Parsing config options and metadata ---------------------------------

    def parse_config_settings(self, config_settings: Optional[Dict]):
        self.runner.verbose = _BuildBackend.is_verbose_enabled(config_settings)

    @staticmethod
    def read_component_config(pyproject_path: Path,
                              config_settings: Optional[Dict],
                              verbose: bool) -> config.ComponentConfig:
        config_settings = config_settings or {}
        try:
            cfg = config.read_component_config(pyproject_path)
            if cfg.dynamic_metadata:
                raise flit_core.config.ConfigError(
                    "Dynamic metadata not supported for components.")
        except flit_core.config.ConfigError as e:
            e.args = ("\n"
                      "\n"
                      "\t❌ Error in user configuration:\n"
                      "\n"
                      f"\t\t{e}\n"
                      "\n", )
            raise
        except AssertionError as e:
            e.args = (
                "\n"
                "\n"
                "\t❌ Internal error while processing the configuration\n"
                "\t   Please notify the developers: https://github.com/tttapa/py-build-cmake/issues\n"
                "\n"
                f"\t\t{e}\n"
                "\n", )
            raise
        if verbose:
            _BuildComponentBackend.print_config_verbose(cfg)
        return cfg

    @staticmethod
    def read_all_metadata(src_dir, config_settings, verbose):
        from flit_core.common import Metadata
        cfg = _BuildComponentBackend.read_component_config(
            src_dir / 'pyproject.toml', config_settings, verbose)
        md_dict = {'name': cfg.package_name}
        md_dict.update(cfg.metadata)
        metadata = Metadata(md_dict)
        metadata.version = _BuildBackend.normalize_version(metadata.version)
        return cfg, metadata

    # --- Building wheels -----------------------------------------------------

    def build_wheel_in_dir(self,
                           wheel_directory_,
                           tmp_build_dir,
                           config_settings,
                           editable=False):
        """This is the main function that contains all steps necessary to build
        a complete wheel package, including the CMake builds etc."""
        comp_source_dir = Path().resolve()

        # Load metadata from the current (component) pyproject.toml file
        comp_cfg, comp_metadata = self.read_all_metadata(
            comp_source_dir, config_settings, self.verbose)

        # Set up all paths
        paths = BuildPaths(
            source_dir=Path(comp_cfg.component['main_project']),
            wheel_dir=Path(wheel_directory_),
            temp_dir=Path(tmp_build_dir),
            staging_dir=Path(tmp_build_dir) / 'staging',
        )

        # Load the config from the main pyproject.toml file
        cfg = _BuildBackend.read_config(paths.source_dir / 'pyproject.toml',
                                        config_settings, self.verbose)

        pkg_info = PackageInfo(
            version=comp_metadata.version,
            package_name=comp_cfg.package_name,
            module_name=cfg.module['name'],  # unused, CMake configuration only
        )

        # Create dist-info folder
        distinfo_dir = f'{pkg_info.package_name}-{pkg_info.version}.dist-info'
        distinfo_dir = paths.staging_dir / distinfo_dir
        os.makedirs(distinfo_dir, exist_ok=True)

        # Write metadata
        metadata_path = distinfo_dir / 'METADATA'
        with open(metadata_path, 'w', encoding='utf-8') as f:
            comp_metadata.write_metadata_file(f)
        # Write or copy license
        _BuildBackend.write_license_files(comp_cfg.license, comp_source_dir,
                                          distinfo_dir)
        # Write entrypoints/scripts
        _BuildBackend.write_entrypoints(distinfo_dir, comp_cfg.entrypoints)

        # Configure, build and install the CMake project
        if cfg.cmake:
            self.do_cmake_build(comp_cfg, paths, cfg, pkg_info)

        # Create wheel
        whl_name = _BuildBackend.create_wheel(paths, cfg, pkg_info)
        return whl_name

    # --- Invoking CMake builds -----------------------------------------------

    def do_cmake_build(self, comp_cfg, paths: BuildPaths, cfg, pkg_info):
        """Configure, build and install using CMake."""

        cmake_cfg, native_cmake_cfg = _BuildBackend.get_cmake_configs(cfg)
        cmaker = _BuildBackend.get_cmaker(paths.source_dir,
                                          paths.staging_dir,
                                          cmake_cfg,
                                          cfg.cross,
                                          None,
                                          pkg_info,
                                          runner=self.runner)

        comp = comp_cfg.component
        if 'build_presets' in comp:
            cmaker.build_settings.presets = comp['build_presets']
        if 'install_presets' in comp:
            cmaker.install_settings.presets = comp['install_presets']
        if 'build_args' in comp:
            cmaker.build_settings.args = comp['build_args']
        if 'build_tool_args' in comp:
            cmaker.build_settings.tool_args = comp['build_tool_args']
        if 'install_args' in comp:
            cmaker.install_settings.args = comp['install_args']
        if 'install_components' in comp:
            cmaker.install_settings.components = comp['install_components']

        if not comp_cfg.component["install_only"]:
            cmaker.build()
        cmaker.install()

    # --- Misc helper functions -----------------------------------------------

    @staticmethod
    def print_config_verbose(cfg):
        print("\npy-build-cmake options")
        print("======================")
        print("component:")
        pprint(cfg.component)
        print("======================\n")


_BACKEND = _BuildComponentBackend()
get_requires_for_build_wheel = _BACKEND.get_requires_for_build_wheel
get_requires_for_build_sdist = _BACKEND.get_requires_for_build_sdist
get_requires_for_build_editable = _BACKEND.get_requires_for_build_editable
build_wheel = _BACKEND.build_wheel
build_sdist = _BACKEND.build_sdist
build_editable = _BACKEND.build_editable
